# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .ogcapi_records_models import (
    Concept,
    Contact1,
    Email,
    Format1,
    Language,
    RecordCommonProperties,
    RecordGeoJSON,
    Theme,
    Type7
)
from .sciencekeywords import (
    KEYWORDS_INDEX,
    ScienceKeywordRecord
)
from ..metadata.software_application_models import (
    AuthorRole,
    CreativeWork,
    DefinedTerm,
    Person,
    SoftwareApplication
)
from ..metadata import Transpiler
from datetime import (
    date,
    datetime,
    timezone
)
from loguru import logger

from pydantic import AnyUrl
from typing import (
    Any,
    Mapping,
    List
)

import os
import time
import uuid

SCIENCE_KEYWORDS_TERM_SET = AnyUrl('https://gcmd.earthdata.nasa.gov/kms/concepts/concept_scheme/sciencekeywords')

DEFAULT_LANGUAGE: Language = Language(
    code='en-US',
    name='English (United States)',
    dir=None
)

def _to_datetime(value: date | datetime):
    if isinstance(value, datetime):
        return value if value.tzinfo else value.replace(tzinfo=timezone.utc)
    return datetime.combine(value, datetime.min.time(), tzinfo=timezone.utc)

def _to_contact(
    author: Person | AuthorRole
) -> Contact1:
    position = None

    if isinstance(author, AuthorRole):
        position = author.role_name
        author = author.author

    return Contact1(
        identifier=str(author.identifier),
        name=f"{author.family_name}, {author.given_name}",
        organization=author.affiliation.name,
        position=position,
        emails=[Email(
            value=author.email
        )]
    )

class OgcRecordsTranspiler(Transpiler):

    def transpile(
        self,
        metadata_source: SoftwareApplication
    ) -> Mapping[str, Any]:
        keywords: List[str] = []
        themes: List[Theme] = []

        if metadata_source.keywords:
            for raw_keyword in metadata_source.keywords if isinstance(metadata_source.keywords, list) else [metadata_source.keywords]:
                if isinstance(raw_keyword, str):
                    keywords.append(raw_keyword)
                elif isinstance(raw_keyword, DefinedTerm):
                    if SCIENCE_KEYWORDS_TERM_SET == raw_keyword.in_defined_term_set and raw_keyword.term_code:
                        if not raw_keyword.term_code in KEYWORDS_INDEX:
                            logger.warning(f"Science Keyword UUID {raw_keyword.term_code} not found in the index")
                        else:
                            science_keyword_record: ScienceKeywordRecord = KEYWORDS_INDEX[str(raw_keyword.term_code)]

                            concepts: List[Concept] = []

                            for i, keyword in enumerate(science_keyword_record.hierarchy_list):
                                concepts.append(
                                    Concept(
                                        id=keyword,
                                        title=None,
                                        description=' > '.join(science_keyword_record.hierarchy_list[0:i+1]),
                                        url=science_keyword_record.uri
                                    )
                                )

                            themes.append(
                                Theme(
                                    scheme=str(SCIENCE_KEYWORDS_TERM_SET),
                                    concepts=concepts
                                )
                            )
                    else:
                        logger.debug(f"Discarding keyword, {raw_keyword}, unsupported")
                else:
                    logger.debug(f"Discarding keyword, {raw_keyword}, unsupported type {type(raw_keyword)}")

        record_geojson: RecordGeoJSON = RecordGeoJSON(
            id=f"urn:uuid:{uuid.uuid4()}",
            type=Type7.FEATURE,
            geometry=None,
            time=None,
            links=None,
            properties=RecordCommonProperties(
                created=_to_datetime(metadata_source.date_created),
                updated=_to_datetime(datetime.fromtimestamp(time.time())),
                title=metadata_source.name,
                description=metadata_source.description if metadata_source.description else None,
                keywords=keywords if keywords else None,
                themes=themes if themes else None,
                language=DEFAULT_LANGUAGE,
                resource_languages=[DEFAULT_LANGUAGE],
                formats=[Format1(
                    name='CWL',
                    media_type='application/cwl'
                )],
                contacts=list(
                    map(
                        _to_contact,
                        metadata_source.author if isinstance(metadata_source.author, list) else [metadata_source.author]
                    )
                ),
                license_=': '.join(
                    list(
                        map(
                            lambda license: str(license.url) if isinstance(license, CreativeWork) else str(license),
                            metadata_source.license if isinstance(metadata_source.license, list) else [metadata_source.license]
                        )
                    )
                )
            )
        )

        return record_geojson.model_dump(by_alias=True, exclude_none=True)



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transpiler_mate.metadata.software_application_models import (
    SoftwareApplication,
    SoftwareSourceCode
)
from .metadata import Transpiler

from giturlparse import parse as gitparse
from loguru import logger
from pydantic import (
    AnyUrl,
    BaseModel
)
from pyld import jsonld
from typing import (
    Any,
    List,
    Mapping,
    MutableMapping
)

class CodeMetaTranspiler(Transpiler):

    def __init__(
        self,
        code_repository: str | None
    ) -> None:
        self.code_repository = code_repository

    def transpile(
        self,
        metadata_source: SoftwareApplication
    ) -> Mapping[str, Any]:
        model: BaseModel = metadata_source

        if self.code_repository:
            parsed_url = gitparse(self.code_repository)

            continuous_integration: str | None = None
            issue_tracker: str | None = None
            related_link: List[str] | None = None

            match parsed_url.platform:
                case 'github':
                    continuous_integration = parsed_url.url2https.replace('.git', '/actions')
                    issue_tracker = parsed_url.url2https.replace('.git', '/issues')
                    related_link = [
                        parsed_url.url2https.replace('.git', page) for page in [
                            '/wiki',
                            '/releases',
                            '/deployments'
                        ]
                    ]

                case 'gitlab':
                    continuous_integration = parsed_url.url2https.replace('.git', '/-/pipelines')
                    issue_tracker = parsed_url.url2https.replace('.git', '/-/issues')
                    related_link = [
                        parsed_url.url2https.replace('.git', page) for page in [
                            '/-/wikis/home',
                            '/-/packages',
                            '/-/pipelines'
                        ]
                    ]

                case _:
                    logger.warning(f"Platform {parsed_url.platform} unsupported yet")

            model = SoftwareSourceCode(
                code_repository=parsed_url.url2https,
                target_product=metadata_source,
                continuous_integration=continuous_integration,
                issue_tracker=issue_tracker,
                related_link=related_link
            ) # type: ignore @type is a constant

        doc: MutableMapping[str, Any] = model.model_dump(
            exclude_none=True,
            by_alias=True
        )

        compacted: MutableMapping[str, Any] = jsonld.compact(
            doc,
            {
                "@context": {
                    "@vocab": "https://schema.org/",
                    # (optional) If you want relative IRIs to stay relative, omit @base.
                    # If you want to forbid a base so @id values don't get resolved, set:
                    # "@base": None,
                }
            },
            options={
                "processingMode": "json-ld-1.1",
                "ordered": None
            }
        ) # type: ignore

        compacted['@context'] = 'https://w3id.org/codemeta/3.0'

        if metadata_source.keywords and isinstance(metadata_source.keywords, list):
            compacted['keywords'] = list(
                filter(
                    lambda keyword: isinstance(keyword, str),
                    metadata_source.keywords
                )
            )

        return compacted



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .metadata import (
    MetadataManager,
    Transpiler
)
from .markdown import markdown_transpile
from datetime import datetime
from enum import (
    auto,
    Enum
)
from functools import wraps
from loguru import logger
from pathlib import Path
from semver import Version
from typing import (
    Tuple,
    Optional
)

import click
import json
import time    

def _track(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()

        logger.info(f"Started at: {datetime.fromtimestamp(start_time).isoformat(timespec='milliseconds')}")

        try:
            func(*args, **kwargs)

            logger.success('------------------------------------------------------------------------')
            logger.success('SUCCESS')
            logger.success('------------------------------------------------------------------------')
        except Exception as e:
            logger.error('------------------------------------------------------------------------')
            logger.error('FAIL')
            logger.error(e)
            logger.error('------------------------------------------------------------------------')

        end_time = time.time()

        logger.info(f"Total time: {end_time - start_time:.4f} seconds")
        logger.info(f"Finished at: {datetime.fromtimestamp(end_time).isoformat(timespec='milliseconds')}")

    return wrapper

@click.group()
def main():
    pass

@main.command(context_settings={'show_default': True})
@click.argument(
    'source',
    type=click.Path(
        path_type=Path,
        exists=True,
        readable=True,
        resolve_path=True
    ),
    required=True
)
@click.option(
    '--base-url',
    type=click.STRING,
    required=True,
    help="The Invenio server base URL"
)
@click.option(
    '--auth-token',
    type=click.STRING,
    required=True,
    envvar='INVENIO_AUTH_TOKEN',
    help="The Invenio Access token"
)
@click.option(
    '--attach',
    type=click.Path(
        path_type=Path,
        exists=True,
        readable=True,
        resolve_path=True
    ),
    multiple=True
)
def invenio_publish(
    source: Path,
    base_url: str,
    auth_token: str,
    attach: Optional[Tuple[Path]]
):
    """
    Publishes the input CWL to an Invenio instance.
    """
    metadata_manager: MetadataManager = MetadataManager(source)

    logger.info(f"Interacting with Invenio server at {base_url})")

    from .invenio import InvenioMetadataTranspiler
    invenio_transpiler: InvenioMetadataTranspiler = InvenioMetadataTranspiler(
        metadata_manager=metadata_manager,
        invenio_base_url=base_url,
        auth_token=auth_token
    )

    record_url = invenio_transpiler.create_or_update_process(
        source=source,
        attach=attach
    )

    logger.success(f"Record available on '{record_url}'")

def _transpile(
    source: Path,
    transpiler: Transpiler,
    output: Path
):
    logger.info(f"Reading metadata from {source}...")
    metadata_manager: MetadataManager = MetadataManager(source)

    logger.success(f"Metadata successfully read!")
    logger.info('Transpiling metadata...')
    data = transpiler.transpile(metadata_manager.metadata)

    logger.success(f"Metadata successfully transpiled!")
    logger.info('Serializing metadata...')
    output.parent.mkdir(parents=True, exist_ok=True)
    with output.open('w') as output_stream:
        json.dump(
            data,
            output_stream,
            indent=2
        )
    
    logger.success(f"Metadata successfully serialized to {output}.")

@main.command(context_settings={'show_default': True})
@click.argument(
    'source',
    type=click.Path(
        path_type=Path,
        exists=True,
        readable=True,
        resolve_path=True
    ),
    required=True
)
@click.option(
    '--code-repository',
    required=False,
    help="The (SVN, GitHub, CodePlex, ...) code repository URL"
)
@click.option(
    '--output',
    type=click.Path(path_type=Path),
    required=False,
    default='codemeta.json',
    help="The output file path"
)
def codemeta(
    source: Path,
    code_repository: str | None,
    output: Path
):
    """
    Transpiles the input CWL to CodeMeta representation.
    """
    from .codemeta import CodeMetaTranspiler
    transpiler: CodeMetaTranspiler = CodeMetaTranspiler(code_repository)

    _transpile(
        source=source,
        transpiler=transpiler,
        output=output
    )

@main.command(context_settings={'show_default': True})
@click.argument(
    'source',
    type=click.Path(
        path_type=Path,
        exists=True,
        readable=True,
        resolve_path=True
    ),
    required=True
)
@click.option(
    '--output',
    type=click.Path(path_type=Path),
    required=False,
    default='record.json',
    help="The output file path"
)
def ogcrecord(
    source: Path,
    output: Path
):
    """
    Transpiles the input CWL to OGC API Record.
    """
    from .ogcapi_records import OgcRecordsTranspiler
    transpiler = OgcRecordsTranspiler()

    _transpile(
        source=source,
        transpiler=transpiler,
        output=output
    )

@main.command(context_settings={'show_default': True})
@click.argument(
    'source',
    type=click.Path(
        path_type=Path,
        exists=True,
        readable=True,
        resolve_path=True
    ),
    required=True
)
@click.option(
    '--output',
    type=click.Path(path_type=Path),
    required=False,
    default='datacite.json',
    help="The output file path"
)
def datacite(
    source: Path,
    output: Path
):
    """
    Transpiles the input CWL to DataCite Metadata.
    """
    from .datacite import DataCiteTranspiler
    transpiler = DataCiteTranspiler()

    _transpile(
        source=source,
        transpiler=transpiler,
        output=output
    )

@main.command(context_settings={'show_default': True})
@click.argument(
    'source',
    type=click.Path(
        path_type=Path,
        exists=True,
        readable=True,
        resolve_path=True
    ),
    required=True
)
@click.option(
    '--workflow-id',
    required=True,
    type=click.STRING,
    default="main",
    help="ID of the main Workflow"
)
@click.option(
    '--output',
    type=click.Path(path_type=Path),
    required=False,
    default='.',
    help="The output directory path"
)
@click.option(
    '--code-repository',
    required=False,
    help="The (SVN, GitHub, CodePlex, ...) code repository URL"
)
def markdown(
    source: Path,
    workflow_id: str,
    output: Path,
    code_repository: str | None
):
    """
    Transpiles the input CWL to Markdown documentation.
    """
    output.mkdir(parents=True, exist_ok=True)

    target: Path = Path(output, f"{workflow_id}.md")

    logger.info(f"Rendering Markdown documentation of {source} to {target}...")

    with target.open("w") as output_stream:
        markdown_transpile(
            source,
            workflow_id,
            output_stream,
            code_repository
        )

    logger.info(f"Markdown documentation successfully serialized to {target}!")


class VersionPart(Enum):
    MAJOR = auto()
    MINOR = auto()
    PATCH = auto()
    BUILD = auto()
    PRE_RELEASE = auto()

@main.command(context_settings={'show_default': True})
@click.argument(
    'source',
    type=click.Path(
        path_type=Path,
        exists=True,
        readable=True,
        resolve_path=True
    ),
    required=True
)
@click.option(
    '--version-part',
    type=click.Choice(
        VersionPart,
        case_sensitive=False
    ),
    required=False,
    default=VersionPart.MINOR,
    help="The version part to update"
)
def bump_version(
    source: Path,
    version_part: VersionPart
):
    """
    Bumps the CWL SW version via SemVer Spec 2.0.0.
    """
    logger.info(f"Reading metadata from {source}...")
    metadata_manager: MetadataManager = MetadataManager(source)

    version = Version.parse(metadata_manager.metadata.software_version)

    if not version.is_valid:
        raise ValueError(f"Version {metadata_manager.metadata.software_version} is not compliant to the Semantic Versioning Specification 2.0.0, see https://semver.org/")

    bumped_version = None

    match version_part:
        case VersionPart.MAJOR:
            bumped_version = version.bump_major()

        case VersionPart.MINOR:
            bumped_version = version.bump_minor()

        case VersionPart.PATCH:
            bumped_version = version.bump_patch()

        case VersionPart.BUILD:
            bumped_version = version.bump_build()

        case VersionPart.PRE_RELEASE:
            bumped_version = version.bump_prerelease()

        case _:
            raise ValueError(f"It's not you, it is us: {version_part} unsupported, but it shouldn't have been happened...")

    logger.success(f"Software Version {metadata_manager.metadata.software_version} updated to {bumped_version}")

    metadata_manager.metadata.software_version = str(bumped_version)

    metadata_manager.update()

for command in [bump_version, codemeta, datacite, invenio_publish, ogcrecord, markdown]:
    command.callback = _track(command.callback)



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cwl_loader import load_cwl_from_yaml
from cwl_loader.utils import search_process
from datetime import datetime
from importlib.metadata import (
    version,
    PackageNotFoundError
)
from jinja2 import (
    Environment,
    PackageLoader
)
from loguru import logger
from pathlib import Path
from transpiler_mate.metadata import MetadataManager
from transpiler_mate.codemeta import CodeMetaTranspiler
from typing import (
    get_args,
    get_origin,
    Any,
    List,
    Mapping,
    TextIO,
    Union
)

import time

# START custom built-in functions to simplify the CWL rendering

# CWLtype to string methods

def type_to_string(typ: Any) -> str:
    '''
    Serializes a CWL type to a human-readable string.

    Args:
        `typ` (`Any`): Any CWL type

    Returns:
        `str`: The human-readable string representing the input CWL type.
    '''
    if get_origin(typ) is Union:
        return " or ".join([type_to_string(inner_type) for inner_type in get_args(typ)])

    if isinstance(typ, list):
        return f"[ {', '.join([type_to_string(t) for t in typ])} ]"

    if hasattr(typ, "items"):
        return f"{type_to_string(typ.items)}[]"

    if isinstance(typ, str):
        return typ

    if hasattr(typ, '__name__'):
        return typ.__name__

    if hasattr(typ, 'type_'):
        return typ.type_
    
    # last hope to follow back
    return str(type)

def _get_version() -> str:
    try:
        return version("transpiler_mate")
    except PackageNotFoundError:
        return 'N/A'

def _to_mapping(
    functions: List[Any]
) -> Mapping[str, Any]:
    mapping: Mapping[str, Any] = {}

    for function in functions:
        mapping[function.__name__] = function

    return mapping

_jinja_environment = Environment(
    loader=PackageLoader(
        package_name='transpiler_mate.markdown'
    )
)
_jinja_environment.filters.update(
    _to_mapping(
        [
            type_to_string
        ]
    )
)
_jinja_environment.tests.update(
    _to_mapping(
        [
            
        ]
    )
)

# END

def markdown_transpile(
    source: Path,
    workflow_id: str,
    output_stream: TextIO,
    code_repository: str | None
):
    logger.info(f"Reading metadata from {source}...")
    metadata_manager: MetadataManager = MetadataManager(source)

    logger.success(f"Metadata successfully read!")
    logger.info('Transpiling metadata...')

    transpiler: CodeMetaTranspiler = CodeMetaTranspiler(code_repository)
    metadata = transpiler.transpile(metadata_manager.metadata)

    logger.success(f"Metadata successfully transpiled!")
    logger.info('Reading Workflow model...')

    cwl_document = load_cwl_from_yaml(metadata_manager.raw_document)

    process = search_process(workflow_id, cwl_document)
    if not process:
        raise ValueError(f"Workflow {workflow_id} does not exist in input CWL document, only {list(map(lambda p: p.id, process)) if isinstance(process, list) else [process.id]} available.")

    logger.success(f"Workflow model successfully read!")

    template = _jinja_environment.get_template(f"workflow.md")

    output_stream.write(
        template.render(
            version=_get_version(),
            timestamp=datetime.fromtimestamp(time.time()).isoformat(timespec='milliseconds'),
            software_source_code=metadata if "SoftwareSourceCode" == metadata["@type"] else None,
            software_application=metadata["targetProduct"] if "SoftwareSourceCode" == metadata["@type"] else metadata,
            workflow=process
        )
    )



# Copyright 2025 Terradue
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .datacite_4_6_models import (
    Affiliation,
    Creator,
    Contributor,
    ContributorType,
    DataCiteAttributes,
    Date,
    DateType,
    Description,
    DescriptionType,
    Identifier,
    NameIdentifier,
    NameType,
    Publisher,
    RelatedIdentifier,
    RelatedIdentifierType,
    RelationType,
    ResourceType,
    ResourceTypeGeneral,
    Right,
    Title
)
from ..metadata import Transpiler
from ..metadata.software_application_models import (
    AuthorRole,
    CreativeWork,
    Organization,
    Person,
    ContributorRole,
    SoftwareApplication
)
from datetime import date
from pydantic import AnyUrl
from typing import (
    Any,
    Mapping
)
from urllib.parse import urlparse

import time
import uuid

__ROLES_MAPPING_: Mapping[AnyUrl, ContributorType] = {
    AnyUrl('http://purl.org/spar/datacite/ContactPerson'): ContributorType.CONTACT_PERSON,
    AnyUrl('http://purl.org/spar/datacite/DataCollector'): ContributorType.DATA_COLLECTOR,
    AnyUrl('http://purl.org/spar/datacite/DataCurator'): ContributorType.DATA_CURATOR,
    AnyUrl('http://purl.org/spar/datacite/DataManager'): ContributorType.DATA_MANAGER,
    AnyUrl('http://purl.org/spar/datacite/Distributor'): ContributorType.DISTRIBUTOR,
    AnyUrl('http://purl.org/spar/datacite/Editor'): ContributorType.EDITOR,
    AnyUrl('http://purl.org/spar/datacite/HostingInstitution'): ContributorType.HOSTING_INSTITUTION,
    AnyUrl('http://purl.org/spar/datacite/Other'): ContributorType.OTHER,
    AnyUrl('http://purl.org/spar/datacite/Producer'): ContributorType.PRODUCER,
    AnyUrl('http://purl.org/spar/datacite/ProjectLeader'): ContributorType.PROJECT_LEADER,
    AnyUrl('http://purl.org/spar/datacite/ProjectManager'): ContributorType.PROJECT_MANAGER,
    AnyUrl('http://purl.org/spar/datacite/ProjectMember'): ContributorType.PROJECT_MEMBER,
    AnyUrl('http://purl.org/spar/datacite/RegistrationAgency'): ContributorType.REGISTRATION_AGENCY,
    AnyUrl('http://purl.org/spar/datacite/RegistrationAuthority'): ContributorType.REGISTRATION_AUTHORITY,
    AnyUrl('http://purl.org/spar/datacite/RelatedPerson'): ContributorType.RELATED_PERSON,
    AnyUrl('http://purl.org/spar/datacite/Researcher'): ContributorType.RESEARCHER,
    AnyUrl('http://purl.org/spar/datacite/ResearchGroup'): ContributorType.RESEARCH_GROUP,
    AnyUrl('http://purl.org/spar/datacite/RightsHolder'): ContributorType.RIGHTS_HOLDER,
    AnyUrl('http://purl.org/spar/datacite/Sponsor'): ContributorType.SPONSOR,
    AnyUrl('http://purl.org/spar/datacite/Supervisor'): ContributorType.SUPERVISOR,
    AnyUrl('http://purl.org/spar/datacite/Translator'): ContributorType.TRANSLATOR,
    AnyUrl('http://purl.org/spar/datacite/WorkPackageLeader'): ContributorType.WORK_PACKAGE_LEADER,
}

def _to_contributor(
    author: Person | ContributorRole
) -> Contributor:
    contributor_type=ContributorType.OTHER

    if isinstance(author, ContributorRole):
        if author.additional_type:
            contributor_type = __ROLES_MAPPING_.get(author.additional_type, ContributorType.OTHER)

        author = author.contributor

    contributor: Contributor = Contributor(
        contributor_type=contributor_type,
        name_type=NameType.PERSONAL,
        name=f"{author.family_name}, {author.given_name}",
        given_name=author.given_name,
        family_name=author.family_name,
    )

    _finalize(
        author=author,
        creator=contributor
    )

    return contributor

def _to_creator(
    author: Person | AuthorRole
) -> Creator:
    if isinstance(author, AuthorRole):
        author = author.author

    creator: Creator = Creator(
        name_type=NameType.PERSONAL,
        name=f"{author.family_name}, {author.given_name}",
        given_name=author.given_name,
        family_name=author.family_name,
    )

    _finalize(
        author=author,
        creator=creator
    )

    return creator

def _finalize(
    author: Person | Organization,
    creator: Creator
):
    if author.identifier:
        creator.name_identifiers = []
        for identifier in author.identifier if isinstance(author.identifier, list) else [author.identifier]:
            scheme, netloc, _, _, _, _ = urlparse(str(identifier))
            creator.name_identifiers.append(
                NameIdentifier(
                    name_identifier=str(identifier),
                    name_identifier_scheme=netloc.split('.')[0].upper(),
                    scheme_uri=f"{scheme}://{netloc}"
                )
            )

    if isinstance(author, Person):
        creator.affiliation = []
        for affiliation in author.affiliation if isinstance(author.affiliation, list) else [author.affiliation]:
            if affiliation.identifier:
                for identifier in affiliation.identifier if isinstance(affiliation.identifier, list) else [affiliation.identifier]:
                    scheme, netloc, _, _, _, _ = urlparse(str(identifier))
                    creator.affiliation.append(
                        Affiliation(
                            affiliation_identifier=str(identifier),
                            affiliation_identifier_scheme=netloc.split('.')[0].upper(),
                            scheme_uri=f"{scheme}://{netloc}"
                        )
                    )

class DataCiteTranspiler(Transpiler):

    def transpile(
        self,
        metadata_source: SoftwareApplication
    ) -> Mapping[str, Any]:
        return DataCiteAttributes(
            doi=metadata_source.identifier,
            types=ResourceType(
                resource_type=metadata_source.name,
                resourceTypeGeneral=ResourceTypeGeneral.SOFTWARE
            ),
            identifiers=[Identifier(
                identifier_type='DOI',
                identifier=metadata_source.identifier
            ) if metadata_source.identifier else Identifier(
                identifier_type='URN',
                identifier=f"urn:uuid:{uuid.uuid4()}"
            )], # supply a fake required identifier if the DOI hasn't been associated yet
            related_identifiers=[RelatedIdentifier(
                related_identifier=str(metadata_source.same_as),
                related_identifier_type=RelatedIdentifierType.DOI,
                relation_type=RelationType.IS_IDENTICAL_TO,
                resource_type_general=ResourceTypeGeneral.SOFTWARE
            )] if metadata_source.same_as else [],
            titles=[Title(
               title= metadata_source.name
            )],
            descriptions=[Description(
                description=metadata_source.description,
                description_type=DescriptionType.TECHNICAL_INFO
            )],
            publisher=Publisher(
                name=metadata_source.publisher.name
            ),
            publication_year=metadata_source.date_created.year,
            dates=[Date(
                date=date.fromtimestamp(time.time()),
                date_type=DateType.UPDATED,
                date_information='New version release'
            )],
            rights=[Right(
                rights=metadata_source.license.name if isinstance(metadata_source.license, CreativeWork) else None,
                rights_uri=metadata_source.license.url if isinstance(metadata_source.license, CreativeWork) else None,
                rights_identifier=metadata_source.license.identifier if isinstance(metadata_source.license, CreativeWork) else None,
                rights_identifier_scheme='SPDX'
            )],
            creators=list(
                map(
                    _to_creator,
                    metadata_source.author if isinstance(metadata_source.author, list) else [metadata_source.author]
                )
            ),
            contributors=list(
                map(
                    _to_contributor,
                    metadata_source.contributor if isinstance(metadata_source.contributor, list) else [metadata_source.contributor]
                )
            ) if metadata_source.contributor else None
        ).model_dump(exclude_none=True, by_alias=True)
